#include <Wire.h>
#include <Adafruit_MCP23017.h>
#include <Adafruit_MCP23008.h>
#include <RotaryEncoder.h>
#include <EEPROM.h>

// Input selector definitions
#define INPUT_SELECTOR_EXPANDER_ADDRESS 7
#define INPUT_SELECTOR_MAX_PINS 8
#define INPUT_SELECTOR_POSITION_MAX 7
#define INPUT_SELECTOR_POSITION_MIN 3
#define INPUT_SELECTOR_POSITION_INVALID 255
#define INPUT_SELECTOR_PINS_PER_INPUT 2
#define INPUT_SELECTOR_SET_PIN 0
#define INPUT_SELECTOR_RESET_PIN -1
#define INPUT_SELECTOR_1 7
#define INPUT_SELECTOR_EEPROM 100
Adafruit_MCP23008 inputSelector;
byte inputSelected;
RotaryEncoder inputEncoder(A8, A9);

// TVC selector definitions
#define TVC_CONTROLLER_COUNT 4
#define TVC_CONTROLLERS_PER_SET 2
#define TVC_MAX_PINS_A 16
#define TVC_MAX_PINS_B 12
#define TVC_MAX_PINS_PER_SET (TVC_MAX_PINS_A + TVC_MAX_PINS_B)
#define TVC_SET_PIN 0
#define TVC_RESET_PIN 1
#define TVC_PINS_PER_TAP 2
#define TVC_POSITION_MAX 26
#define TVC_POSITION_MIN 0
#define TVC_POSITION_INVALID 255
#define TVC_MUTED -1
#define TVC_EEPROM 104
#define TVC_MUTED_EEPROM 108
Adafruit_MCP23017 tvcs[TVC_CONTROLLER_COUNT];
byte tvcSelected;
RotaryEncoder tvcEncoder(A10, A11);

#define RELAY_SETTLE_DELAY 20

// Mute tracking; set to millis when button pushed (LOW); otherwise zero (with small wrap jeopardy)
#define MUTE_BUTTON_PIN 2
#define MUTE_BUTTON_RELEASE_TIME 0
#define MUTE_BUTTON_MUTE_DURATION 50
volatile unsigned long buttonLowTime;
boolean isMuted;

// Reset button; set to millis when button pushed (LOW); otherwise zero (with small wrap jeopardy)
#define RESET_BUTTON_PIN 3
#define RESET_BUTTON_RELEASE_TIME 0
#define RESET_BUTTON_DURATION 10000
volatile unsigned long resetButtonLowTime;

#define PERSISTENT_STORE 1
#define LED_TELLTALE 22

void(* resetFunc) (void) = 0;

/**
 * Sets up the controller to monitor the rotary encoders used for volume adjustment
 * and input selection, then sets up the volume and input subsystems. Finally, refreshes 
 * the display to show the current volume and selected input.
 */
void setup() {
  Serial.begin(115200);
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, LOW);
  pinMode(LED_TELLTALE, OUTPUT);
  digitalWrite(LED_TELLTALE, LOW);

  setupRotaryControllers ();
  setupTVC ();
  tvcEncoder.setPosition (tvcSelected);
  setupInput ();
  setupMuteButton ();
  setupResetButton ();
  setupDisplay (isMuted ? TVC_MUTED : tvcSelected * TVC_PINS_PER_TAP, inputSelected);
  digitalWrite(LED_TELLTALE, HIGH);
}

/** 
 * Interrupt service routine for Pin Change Interrupt 2. These are
 * wired to the two rotary encoders. Calls individual service routines
 * to track encoder changes.
 */
ISR(PCINT2_vect) {
  inputEncoder.tick();
  if (!isMuted)
    tvcEncoder.tick ();
}

/**
 * Main controller loop.
 * 
 * Reads the rotary encoders to determine changes. If no changes have
 * occurred, updates the display as needed. 
 * 
 * If the input rotary encoders has turned in either directions, updates
 * the selected input. Input selection wraps.
 * 
 * If the volume rotary encoders has a new position, updates to the new
 * volume. The volume setting cannot be lower that -50dB or higher than 0dB.
 * Attempted changes outside of these limits are ignored.
 */
void loop() {
  long int now = millis ();
  boolean changeMute = false;
  boolean doReset = false;
  // Read button information with interrupts disabled to ensure atomicity
  noInterrupts ();
  if (buttonLowTime != MUTE_BUTTON_RELEASE_TIME) {
    changeMute = now > buttonLowTime + MUTE_BUTTON_MUTE_DURATION;
    if (changeMute)
      buttonLowTime = MUTE_BUTTON_RELEASE_TIME;
  }
  if (resetButtonLowTime != RESET_BUTTON_RELEASE_TIME) {
    doReset = now > resetButtonLowTime + RESET_BUTTON_DURATION;
    if (doReset)
      resetButtonLowTime = RESET_BUTTON_DURATION;
  }
  interrupts ();
  // Process reset button 
  if (doReset) {
#ifdef PERSISTENT_STORE
    // Invalidate all of the saved information so that after restart, default settings are initialized
    EEPROM.update (TVC_MUTED_EEPROM, false);
    EEPROM.update (INPUT_SELECTOR_EEPROM, INPUT_SELECTOR_POSITION_INVALID);
    EEPROM.update (TVC_EEPROM, TVC_POSITION_INVALID);
#endif
    for (int i = 0; i < 8; i++) {
      digitalWrite(LED_TELLTALE, (i & 1) == 0 ? LOW : HIGH);
      delay (500);
    }
    resetFunc ();
  }
  // Process mute button
  if (changeMute) {
    if (isMuted) {
      // Unmute the volume
      if (tvcSelected != TVC_POSITION_MIN) {
        fireTVCRelay ((tvcSelected * TVC_PINS_PER_TAP) + TVC_SET_PIN);
        fireTVCRelay ((TVC_POSITION_MIN * TVC_PINS_PER_TAP) + TVC_RESET_PIN);
      }
      updateVolume (tvcSelected * TVC_PINS_PER_TAP);
    }
    else {
      // Mute the volume
      if (tvcSelected != TVC_POSITION_MIN) {
        fireTVCRelay ((TVC_POSITION_MIN * TVC_PINS_PER_TAP) + TVC_SET_PIN);
        fireTVCRelay ((tvcSelected * TVC_PINS_PER_TAP) + TVC_RESET_PIN);
      }
      updateVolume (TVC_MUTED);
    }
    isMuted = !isMuted;
#ifdef PERSISTENT_STORE
    EEPROM.update (TVC_MUTED_EEPROM, isMuted);
#endif
  }

  // Read input selection rotary encoder position change if any
  int inputEncoderDirection = inputEncoder.getDirection ();
   // Read TVC tap selection position
  int tvcEncoderPosition = tvcEncoder.getPosition ();
  // If no change, then update the display
  if (inputEncoderDirection == 0 && tvcEncoderPosition == tvcSelected) {
    displayUpdate (isMuted ? TVC_MUTED : tvcSelected * TVC_PINS_PER_TAP, inputSelected);
  }
  else {
    // Handle input selector rotary encoder change
    if (inputEncoderDirection != 0) {
      int newInput = inputSelected + (-inputEncoderDirection * 2);
      if (newInput < INPUT_SELECTOR_POSITION_MIN)
        newInput = INPUT_SELECTOR_POSITION_MAX;
      if (newInput > INPUT_SELECTOR_POSITION_MAX)
        newInput = INPUT_SELECTOR_POSITION_MIN;
      // First, deselect existing input
      fireInputRelay (inputSelected + INPUT_SELECTOR_RESET_PIN);
      // Then, select new input and save it
      fireInputRelay (newInput);
      inputSelected = newInput;
      updateInput (inputSelected);
#ifdef PERSISTENT_STORE
        EEPROM.update (INPUT_SELECTOR_EEPROM, inputSelected);
#endif
    }
    // Handle TVC tap selector position change
    if (tvcEncoderPosition != tvcSelected) {
      int newTVCSelected = tvcEncoderPosition;
      if (newTVCSelected < TVC_POSITION_MIN)
        newTVCSelected = TVC_POSITION_MIN;
      if (newTVCSelected > TVC_POSITION_MAX)
        newTVCSelected = TVC_POSITION_MAX;
      if (newTVCSelected != tvcSelected) {
         // Adjust volume. First select new TVC tap, and then deselect the old one so that there are no popping noises
         fireTVCRelay ((newTVCSelected * TVC_PINS_PER_TAP) + TVC_SET_PIN);
         fireTVCRelay ((tvcSelected * TVC_PINS_PER_TAP) + TVC_RESET_PIN);
         // Save the new volume
         tvcSelected = newTVCSelected;
         updateVolume (tvcSelected * TVC_PINS_PER_TAP);
#ifdef PERSISTENT_STORE
      EEPROM.update (TVC_EEPROM, tvcSelected);
#endif
      }
      else
        tvcEncoder.setPosition (tvcSelected);
    }
  }
}

/**
 * Initializes the rotary encoders for volume control and input
 * selection.
 */
void setupRotaryControllers () {
  // Enable rotary encoder on Pins A8/9 using interrupts on Mega 2560
  *digitalPinToPCMSK(A8) |= bit (digitalPinToPCMSKbit(A8)); 
  *digitalPinToPCMSK(A9) |= bit (digitalPinToPCMSKbit(A9));
  *digitalPinToPCMSK(A10) |= bit (digitalPinToPCMSKbit(A10)); 
  *digitalPinToPCMSK(A11) |= bit (digitalPinToPCMSKbit(A11));
  PCIFR  |= bit (digitalPinToPCICRbit(A8)) | bit (digitalPinToPCICRbit(A9));
  PCICR  |= bit (digitalPinToPCICRbit(A8)) | bit (digitalPinToPCICRbit(A9));
  PCIFR  |= bit (digitalPinToPCICRbit(A10)) | bit (digitalPinToPCICRbit(A10));
  PCICR  |= bit (digitalPinToPCICRbit(A11)) | bit (digitalPinToPCICRbit(A11));
}

/**
 * Initializes the input selector control IO expansion decoder and sets
 * the corresponding relays to a known state if needed.
 * 
 * Currently, only three inputs are implemented in hardware.
 * 
 * Currently, the input is set to the state saved in the EEPROM or input 1 if none 
 * on power up and all other inputs are deselected.
 */
void setupInput () {
  // Initialize the input selector IO expansion decoder
  inputSelector.begin (INPUT_SELECTOR_EXPANDER_ADDRESS);
  // Set  each pin in the decoder to output mode - none are required for input
  for (int i = 0; i < INPUT_SELECTOR_MAX_PINS; i += 1)
    inputSelector.pinMode(i, OUTPUT);
  // initialize currently selected input, if any
#ifdef PERSISTENT_STORE
  inputSelected = EEPROM.read (INPUT_SELECTOR_EEPROM);
#else
  inputSelected = INPUT_SELECTOR_POSITION_INVALID;
#endif
  if (inputSelected > INPUT_SELECTOR_POSITION_MAX) {
    // Set all relays to a known state
    for (int i = INPUT_SELECTOR_POSITION_MAX; i >= INPUT_SELECTOR_POSITION_MIN; i -= INPUT_SELECTOR_PINS_PER_INPUT)
      inputSelector.digitalWrite (i + INPUT_SELECTOR_RESET_PIN, HIGH);
    delay (RELAY_SETTLE_DELAY);
    for (int i = INPUT_SELECTOR_POSITION_MAX; i >= INPUT_SELECTOR_POSITION_MIN; i -= INPUT_SELECTOR_PINS_PER_INPUT)
      inputSelector.digitalWrite (i + INPUT_SELECTOR_RESET_PIN, LOW);
    // Select input 1
    inputSelected = INPUT_SELECTOR_1;
#ifdef PERSISTENT_STORE
    EEPROM.update (INPUT_SELECTOR_EEPROM, inputSelected);
#endif
    fireInputRelay (inputSelected);
  }
}

/**
 * Initializes the transformer volume control IO expansion decoders and sets
 * the corresponding relays to a known state if needed.
 * 
 * Currently, only two decoder chips are implemented.
 * 
 * Currently, The TVC is set to the state saved in the EEPROM or -50dB if none 
 * on power up and all relays are set in sequence due to power limitations.
 */
void setupTVC () {
  // Initialize TVC selector IO expansion decoders
  for (int i = 0; i < TVC_CONTROLLER_COUNT; i += 1)
    tvcs[i].begin (i);
  // Set each pin in the decoder to output mode - none are required for input
  for (int j = 0; j < TVC_CONTROLLER_COUNT; j += 2)
    for (int i = 0; i < TVC_MAX_PINS_A; i += 1)
      tvcs[j].pinMode(i, OUTPUT);
  for (int j = 1; j < TVC_CONTROLLER_COUNT; j += 2)
    for (int i = 0; i < TVC_MAX_PINS_B; i += 1)
      tvcs[j].pinMode(i, OUTPUT);
  // initialize currently selected volume, if any
#ifdef PERSISTENT_STORE
  tvcSelected = EEPROM.read (TVC_EEPROM);
#else
  tvcSelected = TVC_POSITION_INVALID;
#endif
  if (tvcSelected > TVC_POSITION_MAX) {
    // Set volume to minimum position to ensure at least the zero volume tap is selected to avoid signal break
    fireTVCRelay (TVC_POSITION_MIN + TVC_SET_PIN);
    // Set all relays to a known state by running up the volume positions and firing reset pin for each
    for (int i = TVC_POSITION_MIN + TVC_PINS_PER_TAP; i <= TVC_POSITION_MAX * TVC_PINS_PER_TAP; i += TVC_PINS_PER_TAP)
      fireTVCRelay (i + TVC_RESET_PIN);
    tvcSelected = TVC_POSITION_MIN;
#ifdef PERSISTENT_STORE
    EEPROM.update (TVC_EEPROM, tvcSelected);
#endif
    fireTVCRelay (tvcSelected);
  }
}

/**
 * Fires the identified pin for 20ms on the TVC relay set.
 */
void fireTVCRelay (int pin) {
  Adafruit_MCP23017 *i2c;
  int tvcToFire = pin / TVC_MAX_PINS_PER_SET;
  int pinToFire = pin - (tvcToFire * TVC_MAX_PINS_PER_SET);
  boolean isControllerB = pinToFire >= /*(tvcToFire * TVC_MAX_PINS_PER_SET) + */TVC_MAX_PINS_A;
  tvcToFire *= TVC_CONTROLLERS_PER_SET;
  if (isControllerB) {
    tvcToFire += 1;
    pinToFire -= TVC_MAX_PINS_A;
  }
/*
  Serial.print ("fireTVCRelay: ");
  Serial.print ((unsigned int)&tvcs[tvcToFire], HEX);
  Serial.print (" pin: ");
  Serial.print (pin);
  Serial.print (" pin fired: ");
  Serial.println (pinToFire);
*/
  tvcs[tvcToFire].digitalWrite(pinToFire, HIGH);
  delay (RELAY_SETTLE_DELAY);
  tvcs[tvcToFire].digitalWrite(pinToFire, LOW);
  
}

/**
 * Fires the identified pin for 20ms on the input selector relay set.
 */
void fireInputRelay ( int pin) {
/*
  Serial.print ("fireInputRelay pin: ");
  Serial.println (pin);
*/
  inputSelector.digitalWrite(pin, HIGH);
  delay (RELAY_SETTLE_DELAY);
  inputSelector.digitalWrite(pin, LOW);
}

void muteISR () {
  if (digitalRead (MUTE_BUTTON_PIN) == LOW)
    // Button pressed; start duration timer
    buttonLowTime =  millis ();
  else
    // Button Release; cancel timer
    buttonLowTime = MUTE_BUTTON_RELEASE_TIME;
}

void setupMuteButton () {
  pinMode(MUTE_BUTTON_PIN, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(MUTE_BUTTON_PIN), muteISR, CHANGE);
  buttonLowTime = MUTE_BUTTON_RELEASE_TIME;
#ifdef PERSISTENT_STORE
  isMuted = EEPROM.read (TVC_MUTED_EEPROM);
#else
  isMuted = false;
#endif
}

void resetISR () {
  if (digitalRead (RESET_BUTTON_PIN) == LOW)
    // Button pressed; start duration timer
    resetButtonLowTime =  millis ();
  else
    // Button Release; cancel timer
    resetButtonLowTime = RESET_BUTTON_RELEASE_TIME;
}

void setupResetButton () {
  pinMode(RESET_BUTTON_PIN, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(RESET_BUTTON_PIN), resetISR, CHANGE);
  resetButtonLowTime = RESET_BUTTON_RELEASE_TIME;
}
