#include <GxEPD2_BW.h>
#include "Lato_Black38pt7b.h"
#include "Lato_Black16pt7b.h"

#define MAX_DISPAY_BUFFER_SIZE 4736 
#define MAX_HEIGHT(EPD) (EPD::HEIGHT <= MAX_DISPAY_BUFFER_SIZE / (EPD::WIDTH / 8) ? EPD::HEIGHT : MAX_DISPAY_BUFFER_SIZE / (EPD::WIDTH / 8))
GxEPD2_BW<GxEPD2_290, MAX_HEIGHT(GxEPD2_290)> display(GxEPD2_290(/*CS*/ 53, /*DC=*/ 8, /*RST=*/ 9, /*BUSY=*/ 7));

// struct used to pass volume and input to single parameter callbacks
struct VolInp {
  int volume;
  int input;
};

// Display component locations
uint16_t volume_box_x, volume_box_y, volume_box_width, volume_box_height;
uint16_t input_box_x, input_box_y, input_box_width, input_box_height;

// tracking for auto power off after some seconds
// tracking for auto power off after 5 seconds
#define POWER_OFF_DELAY 5000
long poweredOn;
// tracking for volume set display refresh
#define VOLUME_ADJUSTMENT_DELAY 100
long volumeSet;
int newVolume;
// tracking for full screen refresh after 60 seconds after at least 4 partial updates
#define FULL_SCREEN_REFRESH_DELAY 60000
#define MAX_PARTIAL_UPDATES_BEFORE_FULL_UPDATE 4
long poweredOff;
int partialUpdatesSinceFullRefresh;

/**
 * Sets up the display.
 * 
 * Its set up for black graphics on white background, upside down landscape.
 * Draws the first display screen from the initial values.
 */
void setupDisplay(int v, int i) {
  display.init(115200);

  display.setRotation(3);
  display.setTextColor(GxEPD_BLACK);
  volume_box_width = display.width ();
  volume_box_height = display.height () - 48;
  volume_box_x = 0;
  volume_box_y = 0;
  
  input_box_width = display.width ();
  input_box_height = display.height () - volume_box_height - 8;
  input_box_x = 0;
  input_box_y = volume_box_height + 8;

  struct VolInp params;
  params.volume = v;
  params.input = i;
  display.setPartialWindow(0, 0, display.width(), display.height());
  display.drawPaged (updateDisplayCallback, &params);
  display.powerOff();
  poweredOn = 0;
  poweredOff = 0;
  volumeSet = 0;
  partialUpdatesSinceFullRefresh = 0;
}

/**
 * Handles display related timers.
 * 
 * Updates the volume setting on the display after 100ms.
 * 
 * Powers down 5s after a screen update.
 * 
 * Performs a full screen refresh 60s after power down if there
 * have been at least 15 partial updates since the last full screen
 * refresh.
 */
void displayUpdate (int v, int i) {
  unsigned long ticks = millis ();
  if (volumeSet != 0 && ticks > volumeSet + VOLUME_ADJUSTMENT_DELAY) {
    updateVolumeDisplay (newVolume);
    volumeSet = 0;
  }
  else if (poweredOn != 0 && ticks > poweredOn + POWER_OFF_DELAY) {
    display.powerOff ();
    poweredOn = 0;
    if (partialUpdatesSinceFullRefresh > MAX_PARTIAL_UPDATES_BEFORE_FULL_UPDATE)
      poweredOff = ticks;
  }
  else if (poweredOff != 0 && ticks > poweredOff + FULL_SCREEN_REFRESH_DELAY) {
    struct VolInp params;
    params.volume = v;
    params.input = i;
    display.setFullWindow();
    display.drawPaged (updateDisplayCallback, &params);
    display.powerOff();
    poweredOff = 0;
    partialUpdatesSinceFullRefresh = 0;
  }
}

/**
 * Paged callback function to update the entire screen.
 * 
 * Draws the volume and input selection using their own partial callback functions.
 * Also draws the static graphic embellishments that are not refreshed during partial
 * updates.
 */
void updateDisplayCallback (struct VolInp *p) {
  display.fillRect (5, volume_box_height + 1, volume_box_width - 10, 1, GxEPD_BLACK);
  display.fillRect (5, volume_box_height + 1 + 2, volume_box_width - 10, 1, GxEPD_BLACK);
  display.setFont(&Lato_Black16pt7b);
  showInputCallback (p->input);
  display.setFont(&Lato_Black38pt7b);
  showVolumeCallback (p->volume);
}

/**
 * Paged callback function that draws the currently selected input.
 * 
 * Is aware of the strange relay pin sequence for input selection where
 * 7 is Streamer, 5 is Vinyl and 3 is CD
 */
void showInputCallback(int i) {
  uint16_t cursor_y = input_box_y - 6 + input_box_height - 6;
  uint16_t cursor_x;
  char *label;
  switch (i) {
    case 7: // Streamer
      label = "Streamer";
      cursor_x = 5;
      break;
    case 5: // CD
      label = "CD";
      cursor_x = 132;
      break;
    case 3: // Vinyl
      label = "Vinyl";
      cursor_x = 216;
     break;
    default:
      label = "";
      break;
  }
  display.setCursor (cursor_x, cursor_y);
  display.print(label);
}

/**
 * Partially updates the display with the selected input. Uses paged update mode.
 * Note: does not affect areas of the display outside of the input selection area.
 */
void updateInput (int i) {
  Serial.print ("Input set to "); Serial.println (i);
  display.setFont(&Lato_Black16pt7b);
  display.setPartialWindow(input_box_x, input_box_y, input_box_width, input_box_height);
  display.drawPaged(showInputCallback, i);
  partialUpdatesSinceFullRefresh += 1;
  poweredOn = millis ();
  poweredOff = 0;
}

/**
 * Paged callback function that draws the currently volume.
 * 
 * It is aware that the supplied volume value is in the range 0 (lowest) to 50 (highest) volume
 * but displays negative values to correspond with the transformer taps. Negative values indicate
 * that the volume is temporarily muted.
 * 
 * Keeps the volume display horizontally centered.
 */
void showVolumeCallback(int v) {
  uint16_t cursor_y = volume_box_y + 66;
  if (v < 0) {
    uint16_t cursor_x = 54;
    display.setCursor (cursor_x, cursor_y);
    display.print("Mute");
  }
  else {
    int dB = v - 52;
    uint16_t cursor_x = dB == 0 ? 92 : dB > -10 ? 68 : 46;
    display.setCursor (cursor_x, cursor_y);
    display.print(dB);
    display.print("dB");
  }
}

/**
 * Updates the recorded volume and arranges to update the display if there are no
 * additional volume changes within 100ms.
 */
void updateVolume (int v) {
  newVolume = v;
  volumeSet = millis ();
  poweredOff = 0;
}

/**
 * Partially updates the display with the current volume. Uses paged update mode.
 * Note: does not affect areas of the display outside of the input selection area.
 */
void updateVolumeDisplay (int v) {
  Serial.print ("Volume set to "); Serial.println (v);
  display.setFont(&Lato_Black38pt7b);
  display.setPartialWindow(volume_box_x, volume_box_y, volume_box_width, volume_box_height);
  display.drawPaged(showVolumeCallback, v);
  partialUpdatesSinceFullRefresh += 1;
  poweredOn = millis ();
  poweredOff = 0;
}
